/* -----------------------------------------------------------------------------
 K03-21 Prevadec kodovani ceske diakritiky

 Martin Sloup, A04372
 FAV, INIB-INF
 msloup@students.zcu.cz
 arcao@arcao.com
 Soubor file_test.c
----------------------------------------------------------------------------- */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "function.h"
#include "file_test.h"


/* Pocet charset tabulek v poli chars. */
#define CHARS_LENGTH (6)
/* vychozi kodovani, pokud se v textu nebudou vyskytovat ceske znaky */
#define DEFAULT_CHARSET "utf-8"

/**
 * Porovnavaci znaky pro windows-1250, latin2, Kamenicti, iso-8859-2, MacCE a KOI-8 CS2.  UTF-8 a UTF-16 detekujeme jinak.
 * UTF-8 umi rozeznat nativne a UTF-16 podle pocatecniho BOM (Byte Order Mark).
 */
const char *chars[] = {"\x9E\x9A\xE8\xF8\xEF\x9D\xF2\xE1\xE9\xEC\xED\xFD\xF3\xFA\xF9\x8E\x8A\xC8\xD8\xCF\x8D\xD2\xC1\xC9\xCC\xCD\xDD\xD3\xDA\xD9\xA7\xF7\xDF",
											 "\xA7\xE7\x9F\xFD\xD4\x9C\xE5\xA0\x82\xD8\xA1\xEC\xA2\xA3\x85\xA6\xE6\xAC\xFC\xD2\x9B\xD5\xB5\x90\xB7\xD6\xED\xE0\xE9\xDE\xF5\xF6\xE1",
											 "\x91\xA8\x87\xA9\x83\x9F\xA4\xA0\x82\x88\xA1\x98\xA2\xA3\x96\x92\x9B\x80\x9E\x85\x86\xA5\x8F\x90\x89\x8B\x9D\x95\x97\xAC\xAD\xF6\xE1",
											 "\xBE\xB9\xE8\xF8\xEF\xBB\xF2\xE1\xE9\xEC\xED\xFD\xF3\xFA\xF9\xAE\xA9\xC8\xD8\xCF\xAB\xD2\xC1\xC9\xCC\xCD\xDD\xD3\xDA\xD9\xA7\xF7\xDF",
											 "\xEC\xE4\x8B\xDE\x93\xE9\xCB\x87\x8E\x9E\x92\xF9\x97\x9C\xF3\xEB\xE1\x89\xDB\x91\xE8\xC5\xE7\x83\x9D\xEA\xF8\xEE\xF2\xF1\xA4\xD6\xA7",
											 "\xDA\xD3\xC3\xD2\xC4\xD4\xCE\xC1\xD7\xC5\xC9\xD9\xCF\xD5\xCA\xFA\xF3\xE3\xF2\xE4\xF4\xEE\xE1\xF7\xE5\xE9\xF9\xEF\xF5\xEA\x83\xA6\x84"};

/* Nazvy charset souboru, budou se podle nich nacitat tabulky */
char *charsets[] = {"windows-1250", "latin-2", "keybcs2", "iso-8859-2", "mac-ce", "koi-8_cs2", "utf-8", "utf-16"};

/**
 * Funkce vraci 1, pokud je zadany soubor binarni
 */
int is_file_binary(char *filename) {
	FILE *fp;
	int c;
	
	/* otevreme soubor */
	if ((fp = fopen(filename, "rb")) == NULL) {
		error_handler(ENOENT, "is_file_binary::fopen::fp");
		fprintf(stderr, "Soubor '%s' se nepodarilo otevrit pro cteni.\nSystem hlasi: %s.\n", filename, strerror(errno));
		exit(1);
	}

	
	while((c = fgetc(fp)) != EOF) {
		/* nebudeme testovat \0 \t \r \n a znaky > 31 */
		if ((c > 0 && c < 9) || (c > 10 && c < 13) || (c > 13 && c < 32)) {
			fclose(fp);
			/* soubor je binarni, vracime 1 */
			return 1;
		}
	}
	fclose(fp);
	return 0;
}

/**
 * char *detect_encoding(char *filename);
 * Funkce vraci detekovane kodovani souboru filename. Nativne podporuje UTF-8 a
 * to az do velikosti 4 bajtu. Unicode / UTF-16 detekuje na zaklade prvnich dvou
 * znaku, tzv BOM (Byte Order Mark), ktere urcuji poradi ulozeni vyssiho a
 * nissiho bajtu (Big Endian & Little Endian). 
 */
char *detect_encoding(char *filename) {
	/* Par potrebnych promenych. */
	FILE *fp;
	int c;
	unsigned int i, x;

	/* Obecne pro vyhledavani, vyhledavani 8-bit kodovani */
	int diachar_used[CHARS_LENGTH][128];
	int diachar_searched[CHARS_LENGTH];
	int max_searched = 0;
	int probably_charset = -1;
	/* UTF-8 */
	int charset_is_utf8 = 0, charset_is_unicode = 0;
	int utf8_subchars = 0;
	/* UTF-16 */
	int unicode_bom = 0; /* 0 = bigendian, 1 = littleendian */
	int unicode_chars_left = 2;
	
	/* Inicializujeme pole, ktere urcuje zda je cesky znak na urcene ascii hodnote nebo ne */ 
	for (i=0;i<CHARS_LENGTH;i++) for (x=0;x<128;x++) diachar_used[i][x]=0; 
	for (i=0;i<CHARS_LENGTH;i++) for (x=0;x<strlen(chars[i]);x++) diachar_used[i][char2int(chars[i][x]) - 128] = 1; 
	for (i=0;i<CHARS_LENGTH;i++) diachar_searched[i] = 0;
	
	/* nastavime vychozi hodnoty */
	charset_is_utf8 = 1;
	
	unicode_chars_left = 2;
	unicode_bom = 0;
	charset_is_unicode = 0;
	
	/* otevreme soubor */
	if ((fp = fopen(filename, "rb")) == NULL) {
		error_handler(ENOENT, "detect_encoding::fopen::fp");
		fprintf(stderr, "\nCHYBA!!! Soubor '%s' se nepodarilo otevrit pro cteni.\nSystem hlasi: %s.\n", filename, strerror(errno));
		exit(1);
	}
	
	/* cteme znak po znaku */
	while((c = fgetc(fp)) != EOF) {
		c = char2int(c);
		if (c > 127) {
			/* detekce utf-8, testuje se zda odpovida definici znaku v kodovani utf-8 */
			if (charset_is_utf8 == 1) {
				if (utf8_subchars == 0) {
					if (c <= 127) utf8_subchars = 0;
					/* nastavime kolik musi nasledovat dalsich znaku pro hodnoty > 127 */
					else if (c >= 192 && c <= 223) utf8_subchars = 1;
					else if (c >= 224 && c <= 239) utf8_subchars = 2;
					else if (c >= 240 && c <= 247) utf8_subchars = 3;
					else charset_is_utf8 = 0;
				} else if (utf8_subchars == 1 || utf8_subchars == 2 || utf8_subchars == 3) {
					/* testovani tech dalsich znaku, zda odpovidaji forme utf-8, pokud ne, neni to utf-8 a zakazeme testovani utf-8 */
					if (c >= 128 && c <= 191 && utf8_subchars != 0) utf8_subchars--;
					else charset_is_utf8 = 0;
				}
			}
			if (charset_is_unicode == 0 && unicode_chars_left > 0) {
				/* detekce utf-16 / unicode - testuje se na zaklade vyskytu BOM - jiny test neni 100% ucinny */
				if (unicode_chars_left == 2) {
					if (c == 0xFE) unicode_bom = 0;
					if (c == 0xFF) unicode_bom = 1;
					unicode_chars_left--;
				} else if (unicode_chars_left == 1) {
					if (unicode_bom == 0 && c == 0xFF) charset_is_unicode = 1;
					if (unicode_bom == 1 && c == 0xFE) charset_is_unicode = 1;
					unicode_chars_left--;
					if (charset_is_unicode == 1) break;
				} 
			}
			/* pro ostani ascii hledame, zda nacteny znak se nachazi v ceske diakritice */
			for(i=0;i<CHARS_LENGTH;i++) {
				/* pokud ano, pricteme ke kodovani uspesny nalezeny vyskyt */
				if (diachar_used[i][c - 128] == 1) {
					diachar_searched[i] += 1;
				}
			}
		}
	}
	fclose(fp);
	
/** 
 * pokud nebylo detekovano utf-8 a utf-16 (unicode), najdme v kterem kodovani
 * se nachazi nejvice nalezenych vyskytu, a to kodovani prohlasime kodovanim 
 * souboru 
 */
	if (charset_is_utf8 == 0 && charset_is_unicode == 0) {
		for(i=0;i<CHARS_LENGTH;i++) {
			if (max_searched < diachar_searched[i]) {
				max_searched = diachar_searched[i];
				probably_charset = i;
			}
		}
	} else {
		/* jinak prohlasime utf-8 ci utf-16 (unicode) */
		if (charset_is_utf8 == 1) probably_charset = CHARS_LENGTH;
		if (charset_is_unicode == 1) probably_charset = CHARS_LENGTH + 1;
	}
	
/**
 * Bylo-li detekovano kodovani vratime jeho textove jmeno, jinak vratime textove
 * jmeno defaultniho kodovani - utf-8 
 */
	if (probably_charset < 0) {
		return DEFAULT_CHARSET;
	} else {
		return charsets[probably_charset];
	}
}
