/* -----------------------------------------------------------------------------
 K03-21 Prevadec kodovani ceske diakritiky

 Martin Sloup, A04372
 FAV, INIB-INF
 msloup@students.zcu.cz
 arcao@arcao.com
 Soubor function.c
----------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>
#include "function.h"

/**
 * char *my_substr(char *text, int from, int length);
 * Funkce vyrizne kus retezce a vrati ho (reza pomoci kopirovani pameti, tzn. ze
 * alokuje pamet pro vraceny retezec). Retezec vraci vcetne ukoncovaciho znaku
 * '\0'. Funkce pracuje uplne stejne jako PHPckovska fce substr, jen pokud
 * chceme kopirovat az do konce retezce tak se za length dosadi 0. V teto fci
 * funguji i varianty se zapornymi cisly, viz. manual PHP.
 */
char *my_substr(char *text, signed int from, signed int length) {
	char* tmp;
	int str_len;

	str_len = strlen(text);

	/* uprava funkce zapornych hodnot */
	if (from < 0) from = (- from > str_len) ? 0 : str_len + from;
	if (length <= 0) length = (str_len - from + length < 0) ? 0 : str_len - from + ((length == 0) ? 0 : length);
	if (length + from > str_len) length = str_len - from;

	/* alokace potrebne pameti a zkopirovani pozadovaneho mnozstvi znaku */
	if ((tmp = (char *) calloc(length + 2, sizeof(char)))==NULL) error_handler(ENOMEM, "my_substr::malloc");
	memcpy((void *) tmp, text + from, (length) * sizeof(char));
	tmp[length] = '\0';
	return tmp;
}

/**
 * void my_strtr(char *text, char *oldchar, char *newchar);
 * Funkce nahradi kazdy znak uvedeny v oldchar jeho novou variantou v newchar.
 * Zase se jedna o kopii PHP funkce strtr. Jen s rozdilem, ze zmeny provadi
 * primo na zadanem retezci text.
 *
 */
void my_strtr(char *text, char *oldchar, char *newchar) {
	int i,x;

	/* projedeme cely retezec */
	for(i=0; i<strlen(text);i++) {
		/* a projedeme kazdy stary znak */
		for (x=0; x<strlen(oldchar); x++) {
			/* zda se nenachazi na pozici retezce */
			if (text[i]==oldchar[x]) {
				/* pokud ano, nahradime ho novym znakem*/
				text[i] = newchar[x];
				break;
			}
		}
	}
}

/**
 * int my_strrchr(char *text, char s_char);
 * Funkce hleda zadany znak s_char v zadanem textu text. Hledani se provadi
 * odzadu. Pokud je nalezen znak, vrati se jeho pozice (prvni znak ma v retezci
 * odpovida pozici cislo 1!!!). A pokud ne je vracena 0.
 */
int my_strrchr(char *text, char s_char) {
	int i;
	/* najdeme posledni vyskyt znaku, jedeme od konce */
	for(i=strlen(text)-1; i >=0; i--) if (text[i]==s_char) return i+1;
	return 0;
}

/**
 * char **my_strtok(char *text, int item_count);
 * Rozdeli retezec podle mezery nebo podle tabulatoru. Bere v uvahu i vice znaku
 * za sebou. Tzn., ze vice mezer ci tabulatoru ignoruje. Pocet vracenych casti
 * je nutne specifikovat pomoci item_count.
 */
char **my_strtok(char *text, int item_count) {
	int i, x = 0;
	signed int pos_begin = - 1, pos_end = - 1;
	char **tmp;

	/* alokujeme si potrebnych pocet prvku pole - item_count */
	if ((tmp = (char **) malloc(item_count * sizeof(char *)))==NULL) error_handler(ENOMEM, "my_strtok::malloc");

	/* projedeme cely retezec */
	for (i=0;i < strlen(text);i++) {
		/* pokud znak neni mezera nebo tabulator zapamatujeme si jeho pozici */
		if (text[i] == ' ' || text[i] == '\t') {
			continue;
		} else {
			if (pos_begin == -1) pos_begin = i;
			/* pokud dalsi znak jiz je zase mezera nebo tabulator provedeme kopirovani retezce od pocatecni pozice */
			if (text[i+1] == ' ' || text[i+1] == '\t') {
				pos_end = i+1;

				/* alokujeme potrebnou pamet a zkopirujeme potrebny pocet znaku */
				if ((tmp[x] = (char *) malloc(((pos_end - pos_begin) + 1) * sizeof(char)))==NULL) error_handler(ENOMEM, "my_strtok::malloc:1");
				memcpy((void *) tmp[x], (void *) (text + pos_begin), ((pos_end - pos_begin) + 1) * sizeof(char));
				/* nesmime zapomenout na nulovy znak */
				tmp[x][pos_end - pos_begin] = '\0';
				/* pricteme pocet vyskytu */
				x++;
				pos_begin = -1;
				if (x >= item_count) break;
			}
		}
	}
	/* nastavime nenalezene vyskyty v poli na NULL */
	for(i=x;i<item_count;i++) tmp[i] = NULL;
	return tmp;
}

/**
 * char *trim(char *text);
 * Funkce oreze vstupni text (mezeru, \t, \r a \n) a vrati ho jako uplne novy
 * retezec.
 */
char *trim(char *text) {
	int length;
	int begin_pos = -1;
	int end_pos = -1;
	int i;
	char *ret;

	length = strlen(text);
	
	/* najdeme prvni znak, ktery neodpovida mezere ani \t */
	begin_pos = length - 1;
	for(i=0;i<length;i++) {
		if (text[i]!=' ' && text[i]!='\t') {
			begin_pos = i;
			break;
		}
	}
	end_pos = 0;
	/* najdeme znak (od konce), ktery neodpovida mezere ani \t ani \r ci \n */
	for(i=length-1;i>=0;i--) {
		if (text[i]!=' ' && text[i]!='\t' && text[i]!='\r' && text[i]!='\n') {
			end_pos = i + 1;
			break;
		}
	}
	/* osetrime prekrizeni zacatku a konce */
	if (begin_pos > end_pos) {
		new_string(ret, 1);
		return ret;
	}

	/* alokujeme pamet a zkopirujeme potrebny pocet znaku */
	new_string(ret, end_pos - begin_pos + 1);
	memcpy(ret, text + begin_pos, (end_pos - begin_pos) * sizeof(char));
	/* nesmime zapomwnout na nulovy znak */
	ret[end_pos - begin_pos] = '\0';
	return ret;
}

/**
 * char *lcase(char *text);
 * Funkce provede konverzi retezce text na mala pismena
 */ 
char *lcase(char *text) {
	int i;
	for(i=0;i<strlen(text);i++) text[i] = tolower(text[i]);
	return text;
}

/**
 * char *ucase(char *text);
 * Funkce provede konverzi retezce text na velka pismena
 */ 
char *ucase(char *text) {
	int i;
	for(i=0;i<strlen(text);i++) text[i] = toupper(text[i]);
	return text;
}

/* A sorting algorithm, call-compatible with the libc qsort() function.
 *
 * The reason to prefer this to standard qsort() is that quicksort is
 * an unstable sorting algorithm, i.e. the relative ordering of
 * elements with the same comparison value may change. Exactly how the
 * ordering changes depends on implementation specific details like
 * the strategy for choosing the pivot element. Thus a list with
 * "equal" values may be sorted differently between platforms, which
 * potentially can lead to significant differences in the move engine.
 *
 * This is an implementation of the combsort algorithm.
 *
 * Testing shows that it is faster than the GNU libc qsort() function
 * on small data sets and within a factor of two slower for large
 * random data sets. Its performance does not degenerate for common
 * special cases (i.e. sorted or reversed data) but it seems to be
 * susceptible to O(N^2) behavior for repetitive data with specific
 * cycle lengths.
 * 
 * GNU-GO  
 */
void gg_sort(void *base, size_t nel, size_t width,
       int (*cmp)(const void *, const void *))
{
  int gap = nel;
  int swap_made;
  char *end = (char *) base + width * (nel - 1);
  do {
    char *a, *b;
    swap_made = 0;
    gap = (10 * gap + 3) / 13;
    for (a = base, b = a + gap * width; b <= end; a += width, b += width) {
      if (cmp((void *) a, (void *) b) > 0) {
       char *c = a;
       char *d = b;
       size_t size = width;
       while (size-- > 0) {
         char tmp = *c;
         *c++ = *d;
         *d++ = tmp;
       }
       swap_made = 1;
      }
    }
  } while (gap > 1 || swap_made);
}
